#include <sys/types.h>
#include <sys/shm.h>
#include <strings.h>
#include <errno.h>
#include <stdio.h>
#include <sys/wait.h>
#include <unistd.h>

#ifdef DEBUG
#define dprintf printf
#else
#define dprintf if (0) printf
#endif

int
setup ()
{
  key_t mykey;
  int shmid;
  char *testch = NULL;

  mykey = IPC_PRIVATE;

  dprintf ("creating private key...");
  shmid = shmget (mykey, 65536, 0777);
  if (shmid == -1)
    {
      dprintf ("error %d, %s\n", errno, strerror (errno));
      exit (1);
    }
  else
    dprintf ("%0x...", shmid);
  dprintf ("attaching... ");
  testch = (char *) shmat (shmid, NULL, 0);
  if (testch == (char *) -1)
    {
      dprintf ("error %d, %s\n", errno, strerror (errno));
      /* todo: delete the shm area */
      exit (1);
    }
  dprintf ("mapped to %x\n", testch);
  dprintf ("writing \'hello\' ...");
  snprintf (testch, 6, "hello");
  dprintf ("reading ... ");
  dprintf ("'%s'...", testch);
  if (strncmp (testch, "hello", 6))
    {
      dprintf ("failed, string mismatch\n");
      exit (1);
    }
  dprintf ("ok\n");
  return shmid;
}

void
test (int shmid)
{
  char *testch;
  dprintf ("attaching... ");
  testch = (char *) shmat (shmid, NULL, 0);
  if (testch == (char *) -1)
    {
      dprintf ("error %d, %s\n", errno, strerror (errno));
      /* todo: delete the shm area */
      exit (1);
    }
  dprintf ("mapped to %x\n", testch);
/* printf ("writing \'hello\' ...\n");
  snprintf (testch, 6, "hello");*/
  dprintf ("reading ... ");
  dprintf ("'%s'...", testch);
  if (strncmp (testch, "hello", 6))
    {
      dprintf ("failed, string mismatch, got '%s'\n", testch);
      exit (1);
    }
  else
    dprintf ("ok\n");
  exit (0);
}

int
main (int argc, char *argv[])
{
  int shmid;
  int pid;
  int n;
  int status;
  if (argc < 2)
    shmid = setup ();
  else
    {
      dprintf ("ok\n");
      sscanf (argv[1], "%x", &shmid);
      dprintf ("got %0x for shmid\n", shmid);
      test (shmid);
      exit (1);			/* we should not have returned */
    }

  /* fork and exec and then map from the execed process */

  if ((pid = fork ()) == 0)
    {
      char shmidbuf[20];
      dprintf ("forked...\n");
      snprintf (shmidbuf, 20, "%0x", shmid);
      dprintf ("execing (%s %s)...", argv[0], shmidbuf);
      fflush (stdout);
      if (execl (argv[0], argv[0], shmidbuf, NULL))
	{
	  dprintf ("failed %d, %s\n", errno, strerror (errno));
	  exit (1);
	}
    }
/* after fork() */
  if ((n = waitpid (pid, &status, 0)) != pid)
    {
      dprintf ("wait pid failed, pid %d, n %d, errno %d\n", pid, n, errno);
      exit (1);
    }
  else if (WIFEXITED (status))
    {
      /* normal termination - pass child exist code onto our parent */
      exit (WEXITSTATUS (status));
    }
  dprintf ("unexpected return value from waitpid()\n");
  return 1;
}
